import matplotlib.pyplot as plt
import numpy as np


class MultiTaskFourRooms:

    def __init__(self, args):
        layout1 = """\
wwwwwwwwwwwww
w     wg   gw
w     w     w
w           w
w     w     w
w     wg   gw
wwwwwwwwwwwww
"""
        layout2 = """\
wwwwwwwwwwwww
wg    w     w
w     w     w
w           w
w     w     w
w     w     w
ww wwww     w
w     www www
w     w     w
w     w     w
w           w
w     w     w
wwwwwwwwwwwww
"""
        layout3 = """\
wwwwwwwwwwwww
wg   gwg   gw
w     w     w
w           w
w     w     w
wg   gw     w
ww wwwwg   gw
wg   gwww www
w     wg   gw
w     w     w
w           w
wg   gwg   gw
wwwwwwwwwwwww
"""
        layout4 = """\
wwwwwwwwwwwww
wg   gwg   gw
w     w     w
w  g     g  w
w     w     w
wg   gw     w
ww wwwwg   gw
wg   gwww www
w     wg   gw
w  g  w  g  w
w           w
wg   gwg   gw
wwwwwwwwwwwww
"""
        if "layout1" in args:
            layout = layout1
        elif "layout2" in args:
            layout = layout2
        elif "layout3" in args:
            layout = layout3
        elif "layout4" in args:
            layout = layout4
        else:
            raise NotImplementedError
        self.occupancy = np.array([list(map(lambda c: 1 if c == 'w' else 0, line)) for line in layout.splitlines()])
        self.height, self.width = self.occupancy.shape
        self.goal_occupancy = np.array(
            [list(map(lambda c: 1 if c == 'g' else 0, line)) for line in layout.splitlines()])
        # Four possible actions
        # 0: UP
        # 1: DOWN
        # 2: LEFT
        # 3: RIGHT
        self.num_actions = 4
        self.directions = [np.array((-1, 0)), np.array((1, 0)), np.array((0, -1)), np.array((0, 1))]
        self.reward_range = np.array([0, 1])
        self.tostate = {}
        self.goals = []
        self.metadata = None

        statenum = 0
        for i in range(self.height):
            for j in range(self.width):
                if self.occupancy[i, j] == 0:
                    if self.goal_occupancy[i, j] == 1:
                        self.goals.append(statenum)
                    self.tostate[(i, j)] = statenum
                    statenum += 1
        self.num_states = statenum
        self.tocell = {v: k for k, v in self.tostate.items()}
        self.num_tasks = len(self.goals)

        if "random_features" in args:
            self.features = "random"
            tmp = [int(s) for s in args.split() if s.isdigit()]
            assert (tmp.__len__() == 2)
            self.dim = tmp[0]
            self.num_ones = tmp[1]
            self.feature_matrix = np.zeros((self.num_states, self.dim))
            for s in range(self.num_states):
                rng = np.random.RandomState(s)
                one_indices = rng.choice(self.dim, self.num_ones, replace=False)
                self.feature_matrix[s][one_indices] = 1
        elif "tabular_features" in args:
            self.features = "tabular"
            self.dim = self.num_states
            self.feature_matrix = np.zeros((self.num_states, self.num_states))
            for s in range(self.num_states):
                self.feature_matrix[s][s] = 1
        elif "xy_features" in args:
            self.features = "xy"
            self.dim = 2
            self.feature_matrix = np.zeros((self.num_states, 2))
            for s in range(self.num_states):
                cell = self.tocell[s]
                self.feature_matrix[s][0] = cell[0]
                self.feature_matrix[s][1] = cell[1]
        else:
            raise NotImplementedError

        if "tabular_states" in args:
            self.tabular_flag = True
        else:
            self.tabular_flag = False

        self.eps_steps = 0
        self.init_states = list(range(np.sum(self.occupancy == 0)))

    def render(self, show_goal=True):
        current_grid = np.array(self.occupancy)
        current_grid[self.current_cell[0], self.current_cell[1]] = -1
        if show_goal:
            goal_cell = self.tocell[self.goals[self.task_id]]
            current_grid[goal_cell[0], goal_cell[1]] = -1
        return current_grid

    def get_feature_matrix(self):
        return self.feature_matrix

    def render_policy(self, greedy_actions, num_adj_options=2):
        action_num_to_symbol = {}
        for i in range(num_adj_options):
            action_num_to_symbol[i] = str(i+1)
        tmp = ['^', 'v', '<', '>']
        for i in range(num_adj_options, num_adj_options + 4):
            action_num_to_symbol[i] = tmp[i - num_adj_options]
        grid = np.array(self.occupancy).astype(str)
        for i in range(greedy_actions.shape[0]):
            grid[self.tocell[i]] = action_num_to_symbol[greedy_actions[i]]
        grid = np.array2string(grid, separator=',', formatter={'str_kind': lambda x: x})
        return grid

    def render_value(self, values):
        grid = np.array(self.occupancy).astype(str)
        for i in range(values.shape[0]):
            grid[self.tocell[i]] = round(values[i], 1)
        grid = np.array2string(grid, separator=',', formatter={'str_kind': lambda x: x})
        return grid

    def reset(self):
        self.eps_steps = 0
        state = np.random.choice(self.init_states)
        self.task_id = np.random.choice(len(self.goals))
        self.goal_state = self.goals[self.task_id]
        self.current_cell = self.tocell[state]
        obs = self.states_to_obss(state)
        return obs

    def reset_test(self):
        return self.reset()

    def check_available_cells(self, cell):
        available_cells = []

        for action in range(self.action_space.n):
            next_cell = tuple(cell + self.directions[action])

            if not self.occupancy[next_cell]:
                available_cells.append(next_cell)

        return available_cells

    def step(self, action):
        reward = -1
        info = {"restart_without_termination": False}

        next_cell = tuple(self.current_cell + self.directions[action])
        self.prev_cell = self.current_cell
        if not self.occupancy[next_cell]:
            self.current_cell = next_cell

        state = self.tostate[self.current_cell]

        # When goal is reached, it is done
        done = state == self.goal_state

        obs = self.states_to_obss(state)

        self.eps_steps += 1

        return obs, reward, done, info

    def true_model_step(self, states, actions):
        obss = []
        rewards = []
        dones = []
        for i in range(states.shape[0]):
            state = np.where(states[i])[0][0]
            action = actions[i]
            done = state == self.tostate[(3, 1)]
            reward = 1 * done

            next_cell = tuple(self.tocell[state] + self.directions[action])
            if not self.occupancy[next_cell]:
                current_cell = next_cell
            else:
                current_cell = self.tocell[state]

            state = self.tostate[current_cell]

            obs = self.states_to_obss(state)
            obss.append(obs)
            rewards.append(reward)
            dones.append(done)

        return obss, rewards, dones

    def generate_random_obs_action_pairs(self, batch_size):
        states = np.random.choice(self.init_states[0], batch_size, replace=False)
        tmp = np.zeros(self.num_states)
        tmp[states] = 1
        obss = self.states_to_obss(states)
        return np.asarray(obss), np.random.randint(0, self.action_space.n, batch_size)

    def states_to_obss(self, states):
        if self.tabular_flag:
            return states
        else:
            return self.feature_matrix[states]

    def seed(self, seed):
        # numpy already has been seeded in run.py
        pass

    def draw_discovered_options_four_rooms(
            self, exp_name, sweep_id, num_steps, num_options, get_q, get_mu, get_pi,
            get_beta, get_adv_and_interest_adv, get_interest, get_mu_all_tasks
    ):
        q_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        adv_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        interest_adv_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        beta_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        interest_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        mu_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        mu_all_tasks_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        pi_y = [
            [np.zeros((self.width, self.height)) for _ in range(self.num_actions)]
            for _ in range(num_options)
        ]
        pi_x = [
            [np.zeros((self.width, self.height)) for _ in range(self.num_actions)]
            for _ in range(num_options)
        ]
        mu = get_mu()
        mu_all_tasks = get_mu_all_tasks()
        q = get_q()
        adv, interest_adv = get_adv_and_interest_adv()
        pi = get_pi()
        beta = get_beta()
        interest = get_interest()

        num_states = 0
        for option in range(num_options):
            for idx in range(self.num_states):
                i, j = self.tocell[idx]
                if beta_maps[option][i, j] == 0:
                    q_maps[option][i, j] = q[idx, option]
                    adv_maps[option][i, j] = adv[idx, option]
                    interest_adv_maps[option][i, j] = interest_adv[idx, option]
                    beta_maps[option][i, j] = beta[idx, option, 1]
                    interest_maps[option][i, j] = interest[idx, option, 1]
                    mu_maps[option][i, j] = mu[idx, option]
                    mu_all_tasks_maps[option][i, j] = mu_all_tasks[idx, option]
                    num_states += 1
                    for a in range(self.num_actions):
                        y, x = self.directions[a]
                        magnitude = pi[idx, option, a]
                        pi_y[option][a][j, i] = -y * magnitude
                        pi_x[option][a][j, i] = x * magnitude

                else:
                    for a in range(self.num_actions):
                        pi_y[option][a][j, i] = 0
                        pi_x[option][a][j, i] = 0

        rows = ['values', 'adv', 'interest adv', 'pis', 'betas', 'interests', 'mu', 'mu all tasks']
        fig, axs = plt.subplots(len(rows), num_options, figsize=(15, 12))

        # option value
        cell_map = self.occupancy.astype('float64')
        masked_cell_map = np.ma.masked_array(cell_map, cell_map == 0)
        min_q = np.floor(np.min([np.min(q_maps[o]) for o in range(num_options)]))
        max_q = np.ceil(np.max([np.max(q_maps[o]) for o in range(num_options)]))
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('values')]
            else:
                ax = axs[rows.index('values'), o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_q_map = np.ma.masked_array(q_maps[o], cell_map != 0)
            im = ax.imshow(masked_q_map, cmap='Blues')
            ax.axis("off")
            im.set_clim(min_q, max_q)
            ax.title.set_text('q ' + str(o))
        fig.colorbar(im, ax=ax)

        # adv
        cell_map = self.occupancy.astype('float64')
        masked_cell_map = np.ma.masked_array(cell_map, cell_map == 0)
        min_adv = np.floor(np.min([np.min(adv_maps[o]) for o in range(num_options)]))
        max_adv = np.ceil(np.max([np.max(adv_maps[o]) for o in range(num_options)]))
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('adv')]
            else:
                ax = axs[rows.index('adv'), o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_adv_map = np.ma.masked_array(adv_maps[o], cell_map != 0)
            im = ax.imshow(masked_adv_map, cmap='Blues')
            ax.axis("off")
            im.set_clim(min_adv, max_adv)
            ax.title.set_text('adv ' + str(o))
        fig.colorbar(im, ax=ax)

        # interest adv
        cell_map = self.occupancy.astype('float64')
        masked_cell_map = np.ma.masked_array(cell_map, cell_map == 0)
        min_interest_adv = np.floor(np.min([np.min(interest_adv_maps[o]) for o in range(num_options)]))
        max_interest_adv = np.ceil(np.max([np.max(interest_adv_maps[o]) for o in range(num_options)]))
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('interest adv')]
            else:
                ax = axs[rows.index('interest adv'), o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_interest_adv_map = np.ma.masked_array(interest_adv_maps[o], cell_map != 0)
            im = ax.imshow(masked_interest_adv_map, cmap='Blues')
            ax.axis("off")
            im.set_clim(min_interest_adv, max_interest_adv)
            ax.title.set_text('interest adv ' + str(o))
        fig.colorbar(im, ax=ax)

        # pi
        Y = np.arange(0, self.height, 1)
        X = np.arange(0, self.width, 1)
        X, Y = np.meshgrid(X, Y)
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('pis')]
            else:
                ax = axs[rows.index('pis'), o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            for a in range(self.num_actions):
                ax.quiver(X, Y, pi_x[o][a].T, pi_y[o][a].T, scale=15.0)
            ax.axis('off')
            ax.title.set_text('pis ' + str(o))
        fig.colorbar(im, ax=ax)

        # beta
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('betas')]
            else:
                ax = axs[rows.index('betas'), o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_beta_map = np.ma.masked_array(beta_maps[o], cell_map != 0)
            im = ax.imshow(masked_beta_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            ax.title.set_text('betas ' + str(o))
        fig.colorbar(im, ax=ax)

        # interests
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('interests')]
            else:
                ax = axs[rows.index('interests'), o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_interest_map = np.ma.masked_array(interest_maps[o], cell_map != 0)
            im = ax.imshow(masked_interest_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            ax.title.set_text('interests ' + str(o))
        fig.colorbar(im, ax=ax)

        # mu
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('mu')]
            else:
                ax = axs[rows.index('mu'), o]

            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_mu_map = np.ma.masked_array(mu_maps[o], cell_map != 0)
            im = ax.imshow(masked_mu_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            ax.title.set_text('mu ' + str(o))
        fig.colorbar(im, ax=ax)

        # mu over all tasks
        for o in range(num_options):
            if num_options == 1:
                ax = axs[rows.index('mu all tasks')]
            else:
                ax = axs[rows.index('mu all tasks'), o]

            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_mu_all_tasks_map = np.ma.masked_array(mu_all_tasks_maps[o], cell_map != 0)
            im = ax.imshow(masked_mu_all_tasks_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            ax.title.set_text('mu all tasks' + str(o))
        fig.colorbar(im, ax=ax)

        tag = "Discovered_Options"
        plt.savefig("experiment/%s/outputs/%d_%s_%d.pdf" % (exp_name, sweep_id, tag, num_steps))
        plt.close()

    def draw_options_separately(
            self, sweep_id, num_adj_options, get_mu, get_pi,
            get_beta, get_interest, get_mu_all_tasks
    ):
        num_options = num_adj_options + self.num_actions
        beta_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        interest_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        mu_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        mu_all_tasks_maps = [self.occupancy.astype('float64') for _ in range(num_options)]
        pi_y = [
            [np.zeros((self.width, self.height)) for _ in range(self.num_actions)]
            for _ in range(num_options)
        ]
        pi_x = [
            [np.zeros((self.width, self.height)) for _ in range(self.num_actions)]
            for _ in range(num_options)
        ]
        mu = get_mu()
        mu_all_tasks = get_mu_all_tasks()
        pi = get_pi()
        beta = get_beta()
        interest = get_interest()

        num_states = 0
        for option in range(num_options):
            for idx in range(self.num_states):
                i, j = self.tocell[idx]
                if beta_maps[option][i, j] == 0:
                    beta_maps[option][i, j] = beta[idx, option, 1]
                    interest_maps[option][i, j] = interest[idx, option, 1]
                    mu_maps[option][i, j] = mu[idx, option]
                    mu_all_tasks_maps[option][i, j] = mu_all_tasks[idx, option]
                    num_states += 1
                    for a in range(self.num_actions):
                        y, x = self.directions[a]
                        magnitude = pi[idx, option, a]
                        pi_y[option][a][j, i] = -y * magnitude
                        pi_x[option][a][j, i] = x * magnitude

                else:
                    for a in range(self.num_actions):
                        pi_y[option][a][j, i] = 0
                        pi_x[option][a][j, i] = 0

        cell_map = self.occupancy.astype('float64')
        masked_cell_map = np.ma.masked_array(cell_map, cell_map == 0)

        # pi
        fig, axs = plt.subplots(1, num_adj_options, figsize=(3 * num_adj_options, 3))
        Y = np.arange(0, self.height, 1)
        X = np.arange(0, self.width, 1)
        X, Y = np.meshgrid(X, Y)
        for o in range(num_adj_options):
            ax = axs[o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            for a in range(self.num_actions):
                ax.quiver(X, Y, pi_x[o][a].T, pi_y[o][a].T, scale=15.0)
            ax.axis('off')
            # ax.title.set_text('$\pi$' + str(o))

        tag = "pi"
        plt.tight_layout()
        plt.savefig("../plots/%d_%s.pdf" % (sweep_id, tag))
        plt.close()

        # beta
        fig, axs = plt.subplots(1, num_adj_options, figsize=(3 * num_adj_options, 3))
        for o in range(num_adj_options):
            ax = axs[o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_beta_map = np.ma.masked_array(beta_maps[o], cell_map != 0)
            im = ax.imshow(masked_beta_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            # ax.title.set_text('betas ' + str(o))

        tag = "beta"
        plt.tight_layout()
        plt.savefig("../plots/%d_%s.pdf" % (sweep_id, tag))
        plt.close()

        # interests
        fig, axs = plt.subplots(1, num_adj_options, figsize=(3 * num_adj_options, 3))
        for o in range(num_adj_options):
            ax = axs[o]
            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_interest_map = np.ma.masked_array(interest_maps[o], cell_map != 0)
            im = ax.imshow(masked_interest_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            # ax.title.set_text('interests ' + str(o))

        tag = "int"
        plt.tight_layout()
        plt.savefig("../plots/%d_%s.pdf" % (sweep_id, tag))
        plt.close()

        # mu
        fig, axs = plt.subplots(1, num_adj_options, figsize=(3 * num_adj_options, 3))
        for o in range(num_adj_options):
            ax = axs[o]

            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_mu_map = np.ma.masked_array(mu_maps[o], cell_map != 0)
            im = ax.imshow(masked_mu_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            # ax.title.set_text('mu ' + str(o))

        tag = "mu"
        plt.tight_layout()
        plt.savefig("../plots/%d_%s.pdf" % (sweep_id, tag))
        plt.close()

        # mu over all tasks
        fig, axs = plt.subplots(1, num_adj_options, figsize=(3 * num_adj_options, 3))
        for o in range(num_adj_options):
            ax = axs[o]

            im_map = ax.imshow(masked_cell_map, cmap='Greys')
            im_map.set_clim(0, 1)
            masked_mu_all_tasks_map = np.ma.masked_array(mu_all_tasks_maps[o], cell_map != 0)
            im = ax.imshow(masked_mu_all_tasks_map, cmap='Blues')
            ax.axis('off')
            im.set_clim(0, 1)
            # ax.title.set_text('mu all tasks' + str(o))

        tag = "mu_all_tasks"
        plt.tight_layout()
        plt.savefig("../plots/%d_%s.pdf" % (sweep_id, tag))
        plt.close()